"""Provide Ant`s Request, Response, Item and Extractor."""
import typing
import os
from collections.abc import MutableMapping
import tempfile
import webbrowser

from aiohttp import ClientResponse, ClientRequest, hdrs
from aiohttp.typedefs import LooseHeaders
from lxml import html
import ujson

from .exceptions import ItemGetValueError


class Request(ClientRequest):
    def __init__(
        self,
        *args,
        timeout: float = 60,
        response_in_stream: bool = False,
        headers: typing.Optional[LooseHeaders] = None,
        data: typing.Any = None,
        **kwargs,
    ) -> None:
        super().__init__(*args, headers=headers, data=data, **kwargs)

        if headers is None or hdrs.HOST not in headers:
            self.headers.pop(hdrs.HOST)

        self.response_in_stream = response_in_stream
        self.timeout = timeout
        self.data = data


class Response(ClientResponse):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._text = None
        self._html_element = None
        self._json = None

    def get_text(
        self, encoding: typing.Optional[str] = None, errors: str = "strict"
    ) -> str:

        if self._body is None:
            raise ValueError("Read stream first")
        if self._text is None:
            if encoding is None:
                encoding = self.get_encoding()
            self._text = self._body.decode(encoding, errors=errors)
        return self._text

    @property
    def simple_text(self) -> str:
        return self.get_text(errors="ignore")

    def get_json(self, loads: typing.Callable = ujson.loads):
        if self._json is None:
            self._json = loads(self.simple_text)
        return self._json

    @property
    def simple_json(self) -> typing.Any:
        return self.get_json()

    @property
    def html_element(self) -> html.HtmlElement:
        if self._html_element is None:
            self._html_element = html.fromstring(self.simple_text)
        return self._html_element

    def open_in_browser(
        self,
        file_type: str = ".html",
        _open_browser_function: typing.Callable[..., bool] = webbrowser.open,
    ) -> bool:
        fd, path = tempfile.mkstemp(file_type)
        os.write(fd, self._body)
        os.close(fd)
        return _open_browser_function("file://" + path)


class CustomNoneType:
    """Different with "None" obj ("null" in json)
    """

    pass


Item = typing.TypeVar("Item")


def set_value_to_item(item: Item, key: str, value: typing.Any):
    if isinstance(item, MutableMapping):
        item[key] = value
    else:
        setattr(item, key, value)


def get_value_from_item(item: Item, key: str):
    try:
        if isinstance(item, MutableMapping):
            return item[key]
        else:
            return getattr(item, key)
    except (KeyError, AttributeError) as e:
        raise ItemGetValueError from e


class ItemExtractor:
    def __init__(self, item_cls: typing.Type[Item]):
        self.item_cls = item_cls
        self.extractors: typing.Dict[
            str, typing.Callable[[typing.Any], typing.Any]
        ] = dict()

    def add_extractor(
        self, key: str, extractor: typing.Callable[[typing.Any], typing.Any]
    ):
        self.extractors[key] = extractor

    def extract(self, res: Response) -> Item:
        item = self.item_cls()
        for key, extractor in self.extractors.items():
            set_value_to_item(item, key, extractor(res))

        return item


class ItemNestExtractor(ItemExtractor):
    def __init__(
        self,
        item_class: typing.Type[Item],
        root_extractor: typing.Callable[[Response], typing.Sequence],
    ):
        self.root_extractor = root_extractor
        super().__init__(item_class)

    def extract_items(self, res: Response) -> typing.Generator[Item, None, None]:
        for node in self.root_extractor(res):
            yield super().extract(node)


__all__ = [
    "Request",
    "Response",
    "Item",
    "ItemExtractor",
    "ItemNestExtractor",
    "get_value_from_item",
    "set_value_to_item",
]



import asyncio
import os

import pytest
from tenacity import RetryError
import aiohttp
from yarl import URL

from ant_nest.pipelines import Pipeline
from ant_nest.ant import CliAnt, Ant
from ant_nest.things import Request
from ant_nest.exceptions import ThingDropped
from .test_things import fake_response


@pytest.mark.asyncio
async def test_ant():
    class TestPipeline(Pipeline):
        def __init__(self):
            super().__init__()
            self.count = 0

        async def on_spider_open(self):
            pass

        async def on_spider_close(self):
            raise Exception("This exception will be logged")

        def process(self, thing):
            self.count += 1
            return thing

    class TestAnt(Ant):
        item_pipelines = [TestPipeline()]
        request_retries = 0

        async def run(self):
            await self.request("http://test.com")
            await self.collect(object())
            assert self.item_pipelines[0].count == 1

        async def _request(self, req: Request):
            return fake_response(b"")

    ant = TestAnt()
    assert ant.name == "TestAnt"
    await ant.main()

    class Test2Ant(TestAnt):
        async def run(self):
            raise Exception("Test exception")

    await Test2Ant().main()


@pytest.mark.asyncio
async def test_ant_with_retry():
    class Test2Ant(Ant):
        request_retries = 2
        request_retry_delay = 0.1

        def __init__(self):
            super().__init__()
            self.min_retries = 2
            self.retries = 0

        async def run(self):
            return None

        async def _request(self, req: Request):
            if self.retries < self.min_retries:
                self.retries += 1
                raise aiohttp.ClientConnectionError()
            else:
                res = fake_response(b"")
                res.status = 200
                return res

    await Test2Ant().request("https://www.test.com")
    ant = Test2Ant()
    ant.request_retries = 1
    with pytest.raises(RetryError):
        await ant.request("https://www.test.com")
    # with params
    ant.retries = 0
    ant.request_retries = 3
    await ant.request("https://www.test.com")
    ant.retries = 0
    with pytest.raises(aiohttp.ClientConnectionError):
        await ant.request("https://www.test.com", retries=0)

    await ant.main()
    # retry condition

    class Test3Ant(Test2Ant):
        async def _request(self, req: Request):
            if self.retries < self.min_retries:
                self.retries += 1
                res = fake_response(b"")
                res.status = 500
                return res
            else:
                res = fake_response(b"")
                res.status = 200
                return res

    await Test3Ant().request("https://www.test.com")
    ant = Test3Ant()
    ant.request_retries = 1
    with pytest.raises(RetryError):
        await ant.request("https://www.test.com")

    await ant.main()


@pytest.mark.asyncio
async def test_pipelines():
    class TestPipeline(Pipeline):
        async def process(self, thing):
            raise TypeError("Test error")

        def on_spider_open(self):
            raise TypeError("Test error")

        def on_spider_close(self):
            raise TypeError("Test error")

    class TestAnt(Ant):
        item_pipelines = [TestPipeline()]

        async def run(self):
            return None

    pls = [Pipeline() for x in range(10)]
    thing = object()
    ant = TestAnt()
    assert thing is await ant._handle_thing_with_pipelines(thing, pls)
    # with exception
    pls[5] = TestPipeline()
    with pytest.raises(TypeError):
        await ant._handle_thing_with_pipelines(thing, pls)
    # exception will be ignored in "open" and "close" method
    with pytest.raises(TypeError):
        await ant.open()
    with pytest.raises(TypeError):
        await ant.close()


@pytest.mark.asyncio
async def test_ant_report():
    class FakePipeline(Pipeline):
        def process(self, thing: Request):
            if thing.host == "error":
                raise ThingDropped
            return thing

    class FakeAnt(Ant):
        request_pipelines = [FakePipeline()]
        request_retries = 0

        async def run(self):
            pass

        async def _request(self, req: Request):
            return fake_response(b"")

    ant = FakeAnt()
    # request report
    await ant.request("http://test.com")
    assert ant._reports["Request"][1] == 1
    assert ant._reports["Request"][0] == 0
    with pytest.raises(ThingDropped):
        await ant.request("http://error")
    assert ant._drop_reports["Request"][1] == 1

    ant._last_time -= ant._last_time + 1  # report
    ant.report(Request("GET", URL("http://test")))
    assert ant._reports["Request"][0] == 1
    ant.report(Request("GET", URL("http://test")), dropped=True)
    assert ant._drop_reports["Request"][0] == 1

    await ant.main()


@pytest.mark.asyncio
async def test_with_real_request():
    httpbin_base_url = os.getenv("TEST_HTTPBIN", "http://localhost:8080/")

    ant = CliAnt()
    res = await ant.request(httpbin_base_url)
    assert res.status == 200
    # method
    for method in ["GET", "POST", "DELETE", "PUT", "PATCH", "HEAD"]:
        res = await ant.request(httpbin_base_url + "anything", method=method)
        assert res.status == 200
        if method != "HEAD":
            assert res.simple_json["method"] == method
        else:
            assert res.simple_text == ""
        # short way
        res = await ant.request(httpbin_base_url + "anything", method=method)
        assert res.status == 200
        if method != "HEAD":
            assert res.simple_json["method"] == method
        else:
            assert res.simple_text == ""
    # params
    res = await ant.request(httpbin_base_url + "get?k1=v1&k2=v2")
    assert res.status == 200
    assert res.simple_json["args"]["k1"] == "v1"
    assert res.simple_json["args"]["k2"] == "v2"
    res = await ant.request(httpbin_base_url + "get", params={"k1": "v1", "k2": "v2"})
    assert res.status == 200
    assert res.simple_json["args"]["k1"] == "v1"
    assert res.simple_json["args"]["k2"] == "v2"
    # data with str
    res = await ant.request(httpbin_base_url + "post", data="Test data", method="POST")
    assert res.status == 200
    assert res.simple_json["data"] == "Test data"
    # data with dict
    res = await ant.request(httpbin_base_url + "post", method="POST", data={"k1": "v1"})
    assert res.status == 200
    assert res.simple_json["form"]["k1"] == "v1"
    # data with bytes
    res = await ant.request(httpbin_base_url + "post", method="POST", data=b"12345")
    assert res.status == 200
    assert res.simple_json["data"] == "12345"
    # data with file
    with open("tests/test.html", "r") as f:
        file_content = f.read()
        f.seek(0)
        res = await ant.request(httpbin_base_url + "post", method="POST", data=f)
        assert res.status == 200
        assert res.simple_json["data"] == file_content
    # headers
    res = await ant.request(
        httpbin_base_url + "headers", headers={"Custom-Key": "test"}
    )
    assert res.status == 200
    assert res.headers["Content-Type"] == "application/json"
    assert res.simple_json["headers"]["Custom-Key"] == "test"
    # cookies
    res = await ant.request(
        httpbin_base_url + "cookies", cookies={"Custom-Key": "test"}
    )
    assert res.status == 200
    assert res.simple_json["cookies"]["Custom-Key"] == "test"
    # get cookies
    ant.request_allow_redirects = False
    res = await ant.request(httpbin_base_url + "cookies/set?k1=v1&k2=v2")
    assert res.status == 302
    assert res.cookies["k1"].value == "v1"
    assert res.cookies["k2"].value == "v2"
    # redirects
    ant.request_allow_redirects = True
    ant.request_max_redirects = 3
    res = await ant.request(httpbin_base_url + "redirect/2")
    assert res.status == 200
    # with http proxy
    proxy = os.getenv("TEST_HTTP_PROXY", "http://bugman:letmein@localhost:3128")
    ant.request_proxies.append(proxy)
    res = await ant.request("http://httpbin.org/anything")
    assert res.status == 200
    # no proxy anymore
    ant.request_proxies.pop()
    res = await ant.request("http://httpbin.org/anything")
    assert res.status == 200
    # set proxy by request
    res = await ant.request("http://httpbin.org/anything", proxy=proxy)
    assert res.status == 200
    # with stream
    ant.response_in_stream = True
    res = await ant.request(httpbin_base_url + "anything")
    assert res.status == 200
    with pytest.raises(ValueError):
        getattr(res, "simple_text")
    while True:
        chunk = await res.content.read(10)
        if len(chunk) == 0:
            break
    # set streaming by request
    res = await ant.request(httpbin_base_url + "anything", response_in_stream=False)
    assert res.status == 200
    assert res.simple_text is not None

    await ant.close()


@pytest.mark.asyncio
async def test_schedule_task():
    count = 0
    max_count = 10

    async def cor():
        nonlocal count
        count += 1

    ant = CliAnt()

    ant.schedule_tasks((cor() for i in range(max_count)))
    await ant.wait_scheduled_tasks()
    assert count == max_count
    # test with limit
    count = 0
    running_count = 0
    max_running_count = -1
    concurrent_limit = 3

    async def cor():
        nonlocal count, running_count, max_running_count
        running_count += 1
        max_running_count = max(running_count, max_running_count)
        await asyncio.sleep(0.1)
        count += 1
        running_count -= 1

    ant.concurrent_limit = concurrent_limit
    ant.schedule_tasks(cor() for i in range(max_count))
    assert ant.is_running
    await ant.wait_scheduled_tasks()
    assert count == max_count
    assert max_running_count <= concurrent_limit
    # test with exception
    count = 0
    max_count = 3

    async def coro():
        nonlocal count
        count += 1
        raise Exception("Test exception")

    ant.schedule_tasks(coro() for i in range(max_count))
    await ant.wait_scheduled_tasks()
    assert count == max_count

    # test with closed ant
    await ant.close()

    x = coro()
    ant.schedule_task(x)  # this coroutine will not be running
    await ant.close()
    assert count == max_count
    with pytest.raises(Exception):
        await x


@pytest.mark.asyncio
async def test_as_completed():
    ant = CliAnt(loop=asyncio.get_event_loop())
    count = 3

    async def cor(i):
        return i

    right_result = 0
    for c in ant.as_completed((cor(i) for i in range(count)), limit=-1):
        await c
        right_result += 1
    assert right_result == count

    async def cor(i):
        await asyncio.sleep(i * 0.1)
        return i

    right_result = 0  # 0, 1, 2
    for c in ant.as_completed((cor(i) for i in reversed(range(count)))):
        result = await c
        assert result == right_result
        right_result += 1
    assert right_result == count
    # with limit
    right_result = 2  # 2, 1, 0
    for c in ant.as_completed((cor(i) for i in reversed(range(count))), limit=1):
        result = await c
        assert result == right_result
        right_result -= 1
    assert right_result == -1

    await ant.close()


@pytest.mark.asyncio
async def test_as_completed_with_async():

    ant = CliAnt()

    async def cor(x):
        if x < 0:
            raise Exception("Test exception")
        return x

    result_sum = 0
    async for result in ant.as_completed_with_async((cor(i) for i in range(5))):
        result_sum += result
    assert result_sum == sum(range(5))

    result_sum = 0
    async for result in ant.as_completed_with_async(
        (cor(i - 2) for i in range(5)), raise_exception=False
    ):
        result_sum += result
    assert result_sum == sum(range(3))

    async for _ in ant.as_completed_with_async([cor(-1)], raise_exception=False):
        assert _
        raise Exception("This loop should not be entered!")

    with pytest.raises(Exception):
        async for _ in ant.as_completed_with_async([cor(-1)]):
            assert _

    await ant.close()


@pytest.mark.asyncio
async def test_ant_main():
    """Pipeline closed before scheduled coroutines done?"""

    class TestPipeline(Pipeline):
        awake = True

        async def on_spider_close(self):
            self.awake = False

    class TestAnt(Ant):
        item_pipelines = [TestPipeline()]

        in_error = False

        async def run(self):
            self.schedule_task(self.long_cor())

        async def long_cor(self):
            await asyncio.sleep(1)
            self.in_error = not self.item_pipelines[0].awake

    ant = TestAnt()
    await ant.main()
    assert not ant.in_error



import os
import io

import pytest
from yarl import URL
import aiofiles

from ant_nest import pipelines as pls
from ant_nest.things import Request
from ant_nest.exceptions import ThingDropped
from .test_things import fake_response


@pytest.mark.asyncio
async def test_pipeline():
    pl = pls.Pipeline()
    pl.process(Request("GET", URL("https://test.com")))


def test_response_filter_error_pipeline():
    pl = pls.ResponseFilterErrorPipeline()
    res = fake_response(b"")
    err_res = fake_response(b"")
    res.status = 200
    err_res.status = 403
    assert res is pl.process(res)
    with pytest.raises(ThingDropped):
        pl.process(err_res)


def test_request_duplicate_filter_pipeline():
    pl = pls.RequestDuplicateFilterPipeline()
    req = Request("GET", URL("http://test.com"))
    assert pl.process(req) is req
    with pytest.raises(ThingDropped):
        pl.process(req)


def test_item_print_pipeline(item_cls):
    pl = pls.ItemPrintPipeline()
    item = item_cls()
    item.count = 3
    item.info = "hi"
    assert pl.process(item) is item


def test_item_filed_replace_pipeline(item_cls):
    pl = pls.ItemFieldReplacePipeline(["info"])
    item = item_cls()
    item.info = "hi\n,\t\r ant\n"
    pl.process(item)
    assert item.info == "hi, ant"


@pytest.mark.asyncio
async def test_item_base_file_dump_pipeline():
    pl = pls.ItemBaseFileDumpPipeline()
    await pl.dump("/dev/null", "Hello World")
    await pl.dump("/dev/null", b"Hello World")
    await pl.dump("/dev/null", io.StringIO("Hello World"))
    await pl.dump("/dev/null", io.BytesIO(b"Hello World"))
    await pl.dump("/dev/null", open("./tests/test.html"), buffer_size=4)
    async with aiofiles.open("./tests/test.html") as f:
        await pl.dump("/dev/null", f)
    async with aiofiles.open("./tests/test.html", "rb") as f:
        await pl.dump("/dev/null", f, buffer_size=4)

    with pytest.raises(ValueError):
        await pl.dump("/dev/null", None)


@pytest.mark.asyncio
async def test_item_json_dump_pipeline(item_cls):
    pl = pls.ItemJsonDumpPipeline(to_dict=lambda x: x)
    item = item_cls()
    item.count = 1
    assert pl.process(item) is item
    item = item_cls()
    item.info = "hi"
    pl.process(item)
    await pl.on_spider_close()

    # clean file
    ci = os.getenv("TEST_HOST", "localhost")
    if ci == "localhost":
        os.remove("./Item.json")


def test_request_user_agent_pipeline():
    pl = pls.RequestUserAgentPipeline(user_agent="ant")
    req = Request("GET", URL("https://www.hi.com"))
    assert pl.process(req) is req
    assert req.headers["User-Agent"] == "ant"

    req.headers["User-Agent"] = "custom"
    assert pl.process(req).headers["User-Agent"] == "custom"


def test_request_random_user_agent_pipeline():
    pl = pls.RequestRandomUserAgentPipeline()
    req = Request("GET", URL("https://www.hi.com"))
    assert pl.process(req) is req
    assert req.headers.get("User-Agent") is not None

    req.headers["User-Agent"] = "custom"
    assert pl.process(req).headers["User-Agent"] == "custom"

    with pytest.raises(ValueError):
        pls.RequestRandomUserAgentPipeline(system="something")

    with pytest.raises(ValueError):
        pls.RequestRandomUserAgentPipeline(browser="something")

    pl = pls.RequestRandomUserAgentPipeline(system="UnixLike", browser="Firefox")
    user_agent = pl.create()
    assert "X11" in user_agent
    assert "Firefox" in user_agent



import typing
import asyncio
import logging
from collections import defaultdict
import ujson
import os
import random
import re

import aiofiles
from aiohttp.http import SERVER_SOFTWARE
from aiohttp import hdrs

from .things import Response, Request, Item, set_value_to_item, get_value_from_item
from .exceptions import ThingDropped


class Pipeline:
    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    def on_spider_open(self) -> typing.Optional[typing.Awaitable]:
        """Call when ant open, this method can be coroutine function"""

    def on_spider_close(self) -> typing.Optional[typing.Awaitable]:
        """Call when ant close, this method can be coroutine function"""

    def process(self, thing: typing.Any) -> typing.Any:
        """Process things, this method can be coroutine function
        Raise ThingDropped when drop one thing

        :raise ThingDropped
        """
        return thing


# Response pipelines
class ResponseFilterErrorPipeline(Pipeline):
    def process(self, thing: Response) -> Response:
        if thing.status >= 400:
            raise ThingDropped("Response - {:s}".format(str(thing)))
        else:
            return thing


# Request pipelines
class RequestDuplicateFilterPipeline(Pipeline):
    def __init__(self):
        self.__request_urls = set()
        super().__init__()

    def process(self, thing: Request) -> Request:
        if thing.url in self.__request_urls:
            raise ThingDropped("Request duplicate!")
        else:
            self.__request_urls.add(thing.url)
            return thing


class RequestUserAgentPipeline(Pipeline):
    user_agent = (
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 "
        "(KHTML, like Gecko) Chrome/51.0.2704.103 Safari/537.36"
    )

    def __init__(self, user_agent: str = user_agent):
        super().__init__()
        self.user_agent = user_agent

    def process(self, thing: Request) -> Request:
        if thing.headers.get(hdrs.USER_AGENT) == SERVER_SOFTWARE:
            thing.headers[hdrs.USER_AGENT] = self.user_agent
        return thing


class RequestRandomUserAgentPipeline(Pipeline):
    """Create simple and common user agent for request by random,
    It`s easy to add new rule.
    """

    USER_AGENT_FORMAT = "Mozilla/5.0 ({system}) {browser}"
    SYSTEM_FORMATS = {
        "UnixLike": "X11; {unix-like_os} {cpu_type}",
        "MacOS": "Macintosh; Intel Mac OS X {macos_version}",
        "Windows": "Windows NT {windows_version}",
        "Android": "Android {android_version}; Linux",
        "iOS": "{ios_driver}; CPU OS {ios_version} like Mac OS X",
    }
    BROWSER_FORMATS = {
        "Firefox": "Gecko/20100101 Firefox/{firefox_version}",
        "Safari": "AppleWebKit/{webkit_version} (KHTML, like Gecko) "
        "Version/{safari_version} Safari/{safari_version2}",
        "Chrome": "AppleWebKit/{webkit_version} (KHTML, like Gecko) "
        "Chrome/{chrome_version} Safari/{safari_version2}",
    }
    FORMAT_VARS: typing.Dict[str, typing.Tuple[str, ...]] = {
        "unix-like_os": ("Linux", "FreeBSD"),
        "cpu_type": ("x86_64", "i386", "amd64"),
        "macos_version": (
            "10_10_1",
            "10_11_1",
            "10_11_2",
            "10_12_2",
            "10_12_3",
            "10_13_3",
        ),
        "windows_version": ("5_0", "5_1", "5_2", "6_0", "6_1", "10_0"),
        "android_version": ("4_4", "5_0", "5_1", "6_0", "6_1", "7_0", "7_1", "8_0"),
        "ios_driver": ("iPone", "iPod", "iPad"),
        "ios_version": ("6_0", "7_0", "8_1", "9_2", "10_3", "10_2_3", "11_3_3"),
        "firefox_version": ("27.3", "28.0", "31.0", "40.1"),
        "webkit_version": ("533.18.1", "533.19.4", "533.20.25", "534.55.3"),
        "safari_version": ("4.0.1", "5.0.4", "5.1.3", "6.0", "7.0.3"),
        "safari_version2": ("530.19.1", "531.9", "533.16", "533.20.27"),
        "chrome_version": (
            "41.0.2226.0",
            "60.0.1325.223",
            "62.0.1532.123",
            "64.0.3282.119",
        ),
    }

    def __init__(self, system: str = "random", browser: str = "random"):
        if system != "random" and system not in self.SYSTEM_FORMATS.keys():
            raise ValueError("The system {:s} is not supported!".format(system))
        if browser != "random" and browser not in self.BROWSER_FORMATS.keys():
            raise ValueError("The browser {:s} is not supported!".format(browser))

        self.system = system
        self.browser = browser
        super().__init__()

    @staticmethod
    def choice(data: typing.Sequence[str]) -> str:
        return random.choice(data)

    def _format(self, pattern: str) -> str:
        """format system or browser pattern string"""
        kv = {}
        keys = re.findall(r"{(\S+?)}", pattern)

        for key in keys:
            kv[key] = self.choice(self.FORMAT_VARS[key])

        return pattern.format(**kv)

    def create(self) -> str:
        if self.system != "random":
            system_format = self.SYSTEM_FORMATS[self.system]
        else:
            system_format = self.SYSTEM_FORMATS[
                self.choice(list(self.SYSTEM_FORMATS.keys()))
            ]

        if self.browser != "random":
            browser_format = self.BROWSER_FORMATS[self.browser]
        else:
            browser_format = self.BROWSER_FORMATS[
                self.choice(list(self.BROWSER_FORMATS.keys()))
            ]

        return self.USER_AGENT_FORMAT.format(
            system=self._format(system_format), browser=self._format(browser_format)
        )

    def process(self, thing: Request) -> Request:
        if thing.headers.get(hdrs.USER_AGENT) == SERVER_SOFTWARE:
            thing.headers[hdrs.USER_AGENT] = self.create()
        return thing


class RequestRandomComputerUserAgentPipeline(Pipeline):
    SYSTEM_FORMATS = {
        "UnixLike": "X11; {unix-like_os} {cpu_type}",
        "MacOS": "Macintosh; Intel Mac OS X {macos_version}",
        "Windows": "Windows NT {windows_version}",
    }


class RequestRandomMobileUserAgentPipeline(Pipeline):
    SYSTEM_FORMATS = {
        "Android": "Android {android_version}; Linux",
        "iOS": "{ios_driver}; CPU OS {ios_version} like Mac OS X",
    }


# Item pipelines
class ItemPrintPipeline(Pipeline):
    def process(self, thing: Item) -> Item:
        self.logger.info(thing.__repr__())
        return thing


class ItemFieldReplacePipeline(Pipeline):
    """Replace some chars in item`s field string"""

    def __init__(
        self,
        fields: typing.Sequence[str],
        excess_chars: typing.Tuple[str, ...] = ("\r", "\n", "\t"),
    ):
        self.fields = fields
        self.excess_chars = excess_chars
        super().__init__()

    def process(self, thing: Item) -> Item:
        for field in self.fields:
            value: str = get_value_from_item(thing, field)
            for char in self.excess_chars:
                value = value.replace(char, "")
            set_value_to_item(thing, field, value)
        return thing


class ItemBaseFileDumpPipeline(Pipeline):
    @classmethod
    async def dump(
        cls, file_path: str, data: typing.Any, buffer_size: int = 1024 * 1024
    ) -> None:
        """Dump data(binary or text, stream or normal, async or not) to disk file.
        typing.IO data will be closed.
        """
        chunk = None
        if isinstance(data, str):
            file_mode = "w"
        elif isinstance(data, bytes):
            file_mode = "wb"
        elif hasattr(data, "read"):  # readable
            chunk = data.read(buffer_size)
            if asyncio.iscoroutine(chunk):
                chunk = await chunk

            if isinstance(chunk, str):
                file_mode = "w"
            else:
                file_mode = "wb"
        else:
            raise ValueError(
                "The type {:s} is not supported".format(type(data).__class__.__name__)
            )

        async with aiofiles.open(file_path, file_mode) as file:
            if chunk is not None:  # in streaming
                await file.write(chunk)
                while True:
                    chunk = data.read(buffer_size)
                    if asyncio.iscoroutine(chunk):
                        chunk = await chunk

                    if len(chunk) == 0:
                        break
                    else:
                        await file.write(chunk)

                result = data.close()
                if asyncio.iscoroutine(result):
                    await result
            else:
                await file.write(data)


class ItemJsonDumpPipeline(ItemBaseFileDumpPipeline):
    """Dump item to json during pipeline closing"""

    def __init__(
        self, *, to_dict: typing.Callable[[Item], typing.Dict], file_dir: str = "."
    ):
        super().__init__()
        self.file_dir = file_dir
        self.data: typing.DefaultDict[str, typing.List[typing.Dict]] = defaultdict(list)
        self.to_dict = to_dict

    def process(self, thing: Item) -> Item:
        self.data[thing.__class__.__name__].append(self.to_dict(thing))
        return thing

    async def on_spider_close(self):
        for file_name, data in self.data.items():
            data = ujson.dumps(data)
            await self.dump(os.path.join(self.file_dir, file_name + ".json"), data)


__all__ = [var for var in vars().keys() if "Pipeline" in var]



"""setting module for your project"""
import os
import logging
import asyncio

from ant_nest.exceptions import ExceptionFilter

ANT_PACKAGES = ["ants"]
ANT_ENV = os.getenv("ANT_ENV", "production")

if ANT_ENV in ("development", "testing"):
    logging.basicConfig(level=logging.DEBUG)
    asyncio.get_event_loop().set_debug(True)
else:
    logging.basicConfig(level=logging.INFO)
    logging.getLogger().addFilter(ExceptionFilter())

# your ant`s class modules or packages

# custom setting, eg:
# MYSQL_HOST = '127.0.0.1'



from yarl import URL
from ant_nest.ant import Ant
from ant_nest.pipelines import ItemFieldReplacePipeline


class GithubAnt(Ant):
    """Crawl trending repositories from github"""

    item_pipelines = [
        ItemFieldReplacePipeline(
            ("meta_content", "star", "fork"), excess_chars=("\r", "\n", "\t", "  ")
        )
    ]
    concurrent_limit = 1  # save the website`s and your bandwidth!

    async def crawl_repo(self, url):
        """Crawl information from one repo"""
        response = await self.request(url)
        # extract item from response
        item = dict()
        item["title"] = response.html_element.xpath("//h1/strong/a/text()")[0]
        item["author"] = response.html_element.xpath("//h1/span/a/text()")[0]
        item["meta_content"] = "".join(
            response.html_element.xpath(
                '//div[@class="repository-content "]/div[2]//text()'
            )
        )
        item["star"] = response.html_element.xpath(
            '//a[@class="social-count js-social-count"]/text()'
        )[0]
        item["fork"] = response.html_element.xpath('//a[@class="social-count"]/text()')[
            0
        ]
        item["origin_url"] = response.url

        await self.collect(item)  # let item go through pipelines(be cleaned)
        self.logger.info("*" * 70 + "I got one hot repo!\n" + str(item))

    async def run(self):
        """App entrance, our play ground"""
        response = await self.request("https://github.com/explore")
        for url in response.html_element.xpath(
            "/html/body/div[4]/main/div[2]/div/div[2]/div[1]/article/div/div[1]/h1/a[2]/"
            "@href"
        ):
            # crawl many repos with our coroutines pool
            self.schedule_task(self.crawl_repo(response.url.join(URL(url))))
        self.logger.info("Waiting...")



"""CLI entry points"""
import typing
import argparse
import inspect
import os
import sys
import shutil
import asyncio
from importlib import import_module
from pkgutil import iter_modules
import signal
import functools
from asyncio.queues import QueueEmpty
import fnmatch

from .ant import Ant
from . import __version__

__signal_count = 0


def get_ants(paths: typing.List[str]) -> typing.Dict[str, typing.Type[Ant]]:
    """Get ant classes by package path"""
    modules = []
    results = {}
    # get all modules from packages and subpackages
    for path in paths:
        module: typing.Any = import_module(path)
        modules.append(module)
        if hasattr(module, "__path__"):
            package_path = []
            for _, name, ispkg in iter_modules(module.__path__):
                next_path = path + "." + name
                if ispkg:
                    package_path.append(next_path)
                else:
                    modules.append(import_module(next_path))
            if len(package_path) > 0:
                results.update(get_ants(package_path))
    # get and sift ant class obj from modules
    for module in modules:
        for name, obj in inspect.getmembers(module):
            if isinstance(obj, type) and issubclass(obj, Ant) and obj is not Ant:
                results[module.__name__ + "." + obj.__name__] = obj
    return results


def shutdown_ant(ants: typing.List[Ant]):
    global __signal_count
    ant_names = "+".join([ant.name for ant in ants])

    if __signal_count == 1:
        print(
            "Receive shutdown command twice, ant {:s} shutdown "
            "immediately".format(ant_names)
        )
        sys.exit()
    __signal_count += 1

    print("Graceful shutdown {:s}...Try again to force " "shutdown".format(ant_names))

    # drop waiting coroutines
    for ant in ants:
        ant._is_closed = True
        while True:
            try:
                ant._queue.get_nowait()
            except QueueEmpty:
                break


def main(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--ants", help="ant names, multi name split by space")
    parser.add_argument("-l", "--list", help="list ants", action="store_true")
    parser.add_argument(
        "-v", "--version", help="get package version", action="store_true"
    )
    parser.add_argument("-c", "--project", help="project name")
    args = parser.parse_args(args)
    sys.path.append(os.getcwd())

    if args.version:
        print(__version__)
        exit()
    elif args.project:
        from . import _settings_example

        try:
            os.mkdir(args.project)
        except FileExistsError:  # pragma: no cover
            raise Exception(f"The project {args.project} exists!")
        os.mkdir(os.path.join(args.project, "ants"))
        shutil.copyfile(
            _settings_example.__file__, os.path.join(args.project, "settings.py")
        )
        exit()
    # in one project
    try:
        import settings
    except ImportError as e:
        print(e)
        print('Are "settings.py" created?')
        exit(-1)
    try:
        ants = get_ants(settings.ANT_PACKAGES)
    except Exception as e:
        print("There is a problem with finding and loading ants:")
        raise e

    if args.list:
        if len(ants) == 0:
            print("Can`t find any ant from " + ",".join(settings.ANT_PACKAGES))
            exit(-1)
        else:
            print("\n".join(ants.keys()))
    elif args.ants is not None:
        selected_ants: typing.List[Ant] = []
        for name in args.ants.split("+"):
            temp = fnmatch.filter(ants.keys(), name)
            if len(temp) == 0:
                print('Can not find ant by the name "{:s}"'.format(name))
                exit(-1)
            else:
                selected_ants.extend([ants[k]() for k in temp])

        loop = asyncio.get_event_loop()
        loop.add_signal_handler(
            signal.SIGINT, functools.partial(shutdown_ant, selected_ants)
        )
        loop.add_signal_handler(
            signal.SIGTERM, functools.partial(shutdown_ant, selected_ants)
        )
        loop.run_until_complete(asyncio.gather(*(ant.main() for ant in selected_ants)))


if __name__ == "__main__":  # pragma: no cover
    main()



from yarl import URL
from ant_nest.ant import Ant
from ant_nest.pipelines import ItemFieldReplacePipeline
from ant_nest.things import ItemExtractor


class GithubAnt(Ant):
    """Crawl trending repositories from github"""

    item_pipelines = [
        ItemFieldReplacePipeline(
            ("meta_content", "star", "fork"), excess_chars=("\r", "\n", "\t", "  ")
        )
    ]
    concurrent_limit = 1  # save the website`s and your bandwidth!

    def __init__(self):
        super().__init__()
        self.item_extractor = ItemExtractor(dict)
        self.item_extractor.add_extractor(
            "title", lambda x: x.html_element.xpath("//h1/strong/a/text()")[0]
        )
        self.item_extractor.add_extractor(
            "author", lambda x: x.html_element.xpath("//h1/span/a/text()")[0]
        )
        self.item_extractor.add_extractor(
            "meta_content",
            lambda x: "".join(
                x.html_element.xpath(
                    '//div[@class="repository-content "]/div[2]//text()'
                )
            ),
        )
        self.item_extractor.add_extractor(
            "star",
            lambda x: x.html_element.xpath(
                '//a[@class="social-count js-social-count"]/text()'
            )[0],
        )
        self.item_extractor.add_extractor(
            "fork",
            lambda x: x.html_element.xpath('//a[@class="social-count"]/text()')[0],
        )
        self.item_extractor.add_extractor("origin_url", lambda x: str(x.url))

    async def crawl_repo(self, url):
        """Crawl information from one repo"""
        response = await self.request(url)
        # extract item from response
        item = self.item_extractor.extract(response)
        item["origin_url"] = response.url

        await self.collect(item)  # let item go through pipelines(be cleaned)
        self.logger.info("*" * 70 + "I got one hot repo!\n" + str(item))

    async def run(self):
        """App entrance, our play ground"""
        response = await self.request("https://github.com/explore")
        for url in response.html_element.xpath(
            "/html/body/div[4]/main/div[2]/div/div[2]/div[1]/article/div/div[1]/h1/a[2]/"
            "@href"
        ):
            # crawl many repos with our coroutines pool
            self.schedule_task(self.crawl_repo(response.url.join(URL(url))))
        self.logger.info("Waiting...")



import sys
import asyncio
from unittest import mock
import re

from yarl import URL
import pytest
import jpath
from lxml import html

from ant_nest.ant import CliAnt
from ant_nest.cli import get_ants
from ant_nest.things import (
    Response,
    Request,
    ItemExtractor,
    set_value_to_item,
    get_value_from_item,
    ItemNestExtractor,
)
from ant_nest.exceptions import ThingDropped, ItemGetValueError, ExceptionFilter
from ant_nest import cli


def fake_response(content):
    res = Response(
        "GET",
        URL("http://test.com"),
        writer=None,
        continue100=None,
        timer=None,
        request_info=None,
        traces=None,
        loop=asyncio.get_event_loop(),
        session=None,
    )
    res._content = content
    res._body = content

    return res


def test_request():
    req = Request("GET", URL("http://test.com"))
    assert req.method == "GET"
    req.__repr__()


def test_response():
    res = fake_response(b"1")
    assert res.get_text(encoding="utf-8") == "1"
    assert res.simple_text == "1"
    assert res.simple_json == 1

    res = fake_response(None)
    with pytest.raises(ValueError):
        res.get_text()
    res = fake_response(b"1")
    res.get_encoding = lambda: "utf-8"
    res._get_encoding = lambda: "utf-8"
    assert res.get_text() == "1"

    res = fake_response(b"")

    def open_browser_function(url):
        return True

    assert res.open_in_browser(_open_browser_function=open_browser_function)


def test_set_get_item():
    class ClsItem:
        pass

    for item in (dict(), ClsItem()):
        set_value_to_item(item, "name", "test")
        assert get_value_from_item(item, "name") == "test"
        with pytest.raises(ItemGetValueError):
            get_value_from_item(item, "name2")


def test_extract_item():
    with open("./tests/test.html", "rb") as f:
        response = fake_response(f.read())
        response.get_text(encoding="utf-8")

    class Item:
        pass

    # extract item with xpath and regex
    item_extractor = ItemExtractor(Item)
    item_extractor.add_extractor(
        "paragraph", lambda x: x.html_element.xpath("/html/body/div/p/text()")[0]
    )
    item_extractor.add_extractor(
        "title", lambda x: re.findall(r"<title>([A-Z a-z]+)</title>", x.simple_text)[0]
    )
    item = item_extractor.extract(response)
    assert item.paragraph == "test"
    assert item.title == "Test html"
    # extract with jpath
    response = fake_response(b'{"a": {"b": {"c": 1}}, "d": null}')
    response.get_text(encoding="utf-8")
    item_extractor = ItemExtractor(Item)
    item_extractor.add_extractor(
        "author", lambda x: jpath.get_all("a.b.c", x.simple_json)[0]
    )
    item_extractor.add_extractor(
        "freedom", lambda x: jpath.get_all("d", x.simple_json)[0]
    )
    item = item_extractor.extract(response)
    assert item.author == 1
    assert item.freedom is None
    # ItemNestExtractor tests
    with open("./tests/test.html", "rb") as f:
        response = fake_response(f.read())
        response.get_text(encoding="utf-8")
    item_nest_extractor = ItemNestExtractor(
        Item, lambda x: x.html_element.xpath('//div[@id="nest"]/div')
    )
    item_nest_extractor.add_extractor("xpath_key", lambda x: x.xpath("./p/text()")[0])
    item_nest_extractor.add_extractor(
        "regex_key",
        lambda x: re.findall(r"regex(\d+)</", html.tostring(x, encoding="unicode"))[0],
    )
    temp = 1
    for item in item_nest_extractor.extract_items(response):
        assert item.xpath_key == str(temp)
        assert item.regex_key == str(temp)
        temp += 1


def test_cli_get_ants():
    ants = get_ants(["ant_nest", "tests"])
    assert CliAnt is list(ants.values())[0]


def test_cli_shutdown():
    ant = CliAnt()
    ant._queue.put_nowait(object())
    cli.shutdown_ant([ant])
    assert ant._is_closed
    assert ant._queue.qsize() == 0

    with pytest.raises(SystemExit):
        cli.shutdown_ant([ant])


def test_cli():

    with pytest.raises(SystemExit):
        cli.main(["-v"])

    with pytest.raises(SystemExit):  # no settings.py
        cli.main(["-l"])

    from ant_nest import _settings_example as settings

    # mock settings.py import
    sys.modules["settings"] = settings

    settings.ANT_PACKAGES = ["NoAnts"]
    with pytest.raises(ModuleNotFoundError):  # can`t import NoAnts
        cli.main(["-l"])

    settings.ANT_PACKAGES = ["ant_nest.things"]
    with pytest.raises(SystemExit):  # no ants can be found
        cli.main(["-l"])

    settings.ANT_PACKAGES = ["tests"]
    cli.main(["-l"])

    with pytest.raises(SystemExit):  # FakeAnt not exist
        cli.main(["-a" "FakeAnt"])
        cli.main(["-l"])

    with pytest.raises(SystemExit), mock.patch("os.mkdir", lambda x: None), mock.patch(
        "shutil.copyfile", lambda *args: None
    ):
        cli.main(["-c" "."])

    cli.main(["-a" "tests.test_things.CliAnt"])


def test_exception_filter():
    class FakeRecord:
        pass

    fr = ExceptionFilter()
    rc = FakeRecord()
    rc.exc_info = (ThingDropped, None, None)

    assert not fr.filter(rc)
    rc.exc_info = (OSError, None, None)
    assert fr.filter(rc)



#!/usr/bin/env python3
from setuptools import setup, find_packages
import re

with open("ant_nest/__init__.py", "rt", encoding="utf8") as f:
    match = re.search(r"__version__ = \"(.*?)\"", f.read())
    if match:
        version = match.group(1)
    else:
        raise Exception("Version not found!")

requires = [
    "aiohttp>=3.3.0",
    "tenacity>=4.8.0",
    "ujson>=1.3.4",
    "aiofiles>=0.3.1",
    "lxml>=3.7.0",
    "typing_extensions>=3.6",
]
tests_requires = [
    "pytest>=3.3.1",
    "pytest-asyncio>=0.8.0",
    "pytest-cov>=2.5.1",
    "jpath>=1.6",
]
setup_requires = ["pytest-runner>=3.0"]

setup(
    name="ant_nest",
    version=version,
    url="https://github.com/strongbugman/ant_nest",
    description="A simple and clear Web Crawler framework build on python3.6+ "
    "with async",
    long_description=open("README.rst").read(),
    author="Bruce Wu",
    author_email="strongbugman@gmail.com",
    license="LGPL",
    classifiers=[
        "Environment :: Console",
        "Programming Language :: Python :: 3.6",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    packages=find_packages(exclude=("tests", "tests.*")),
    install_requires=requires,
    entry_points={"console_scripts": ["ant_nest = ant_nest.cli:main"]},
    setup_requires=setup_requires,
    tests_require=tests_requires,
)
